import numpy as np
import birdepy.utility as ut
from birdepy.interface_probability import probability
import warnings


def bld_ll_fun(data, likelihood, model, z_trunc, options):
    if likelihood in ['da', 'gwa', 'gwasa', 'ilt', 'oua']:
        def ll_fun(param):
            ll = 0
            for i in data:
                pr = probability(i[0], i[1], i[2], param, method=likelihood,
                                 model=model, z_trunc=z_trunc,
                                 options=options)[0][0]
                if np.isnan(pr):
                    warnings.warn("Computation of a transition probability has failed and been "
                                  "replaced by 1e-100. The results may be unreliable. ",
                                  category=RuntimeWarning)
                    ll += data[i] * np.log(1e-100)
                else:
                    ll += data[i] * np.log(np.maximum(1e-100, pr))
            return ll
    elif likelihood in ['Erlang', 'expm', 'uniform']:
        sorted_data = ut.data_sort_2(data)
        z_min, z_max = z_trunc

        def ll_fun(param):
            ll = 0
            for t in sorted_data:
                p_mat = probability(np.arange(z_min, z_max + 1, 1),
                                    np.arange(z_min, z_max+1, 1), t, param,
                                    method=likelihood, model=model,
                                    z_trunc=z_trunc, options=options)
                for i in sorted_data[t]:
                    ll += sorted_data[t][i] * np.log(
                        np.maximum(1e-100, p_mat[i[0] - z_min, i[1] - z_min]))
            return ll

    else:
        raise TypeError("Argument likelihood has an unknown value.")

    return ll_fun


def discrete_est_dnm(data, likelihood, model, z_trunc, idx_known_p, known_p,
                     p0, p_bounds, con, opt_method, options):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using direct numerical maximization of
    approximate likelihood functions.
    See :ref:`here <Direct Numerical Maximization>` for more information.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'dnm'::

        birdepy.estimate(t_data, p_data, p0, p_bounds, framework='dnm', likelihood='expm', z_trunc=())

    The parameters associated with this framework (listed below) can be
    accessed using kwargs in :func:`birdepy.estimate()`. See documentation of
    :func:`birdepy.estimate` for the main arguments.

    Parameters
    ----------
    likelihood : string, optional
        Likelihood approximation method. Should be one of
        (alphabetical ordering):

            - 'da' (see :ref:`here <birdepy.probability(method='da')>`)
            - 'Erlang' (see :ref:`here <birdepy.probability(method='Erlang')>`)
            - 'expm'  (default) (see :ref:`here <birdepy.probability(method='expm')>`)
            - 'gwa' (see :ref:`here <birdepy.probability(method='gwa')>`)
            - 'gwasa' (see :ref:`here <birdepy.probability(method='gwasa')>`)
            - 'ilt' (see :ref:`here <birdepy.probability(method='ilt')>`)
            - 'oua' (see :ref:`here <birdepy.probability(method='oua')>`)
            - 'uniform' (see :ref:`here <birdepy.probability(method='uniform')>`)
        The links point to the documentation of the relevant `method` in
        :func:`birdepy.probability`. The arguments associated with each of
        these methods may be used as a kwarg in :func:`birdepy.estimate()`
        when `likelihood` is set to use the method.

    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where
        ``z_min=max(0, min(p_data) - 2*z_range)``
        and ``z_max=max(p_data) + 2*z_range`` where
        ``z_range=max(p_data)-min(p_data)``. Only applicable to `likelihood`
        methods 'Erlang', 'expm' and 'uniform'.

    Examples
    --------
    Simulate a sample path and estimate the parameters using the various
    likelihood approximation methods.

    The constraint ``con={'type': 'ineq', 'fun': lambda p: p[0]-p[1]}`` ensures
    that p[0] > p[1] (i.e., rate of spread greater than recovery rate).

    >>> import birdepy as bd
    >>> t_data = [t for t in range(100)]
    >>> p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
    ...                               survival=True, seed=2021)
    >>> for likelihood in ['da', 'Erlang', 'expm', 'gwa', 'gwasa', 'ilt', 'oua', 'uniform']:
    ...     likelihood = 'gwasa'
    >>>     est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[1e-6,1], [1e-6,1], [1e-6, 0.1]],
    ...                       framework='dnm', model='Ricker', idx_known_p=[3], known_p=[1],
    ...                       likelihood=likelihood)
    >>>     print('dnm estimate using', likelihood, 'is', est.p, ', with standard errors',
    ...           est.se, 'computed in ', est.compute_time, 'seconds.')
    dnm estimate using da is [0.742967758293063, 0.21582082476775125, 0.022598554340938885] , with standard errors [0.16633436 0.03430501 0.00436483] computed in  12.29400086402893 seconds.
    dnm estimate using Erlang is [0.7478158070214841, 0.2150341741826537, 0.022748822721356705] , with standard errors [0.16991822 0.0345762  0.00435054] computed in  0.9690003395080566 seconds.
    dnm estimate using expm is [0.7477255598292176, 0.2150476994316206, 0.022745305129350565] , with standard errors [0.16904095 0.03443197 0.00433563] computed in  1.6919987201690674 seconds.
    dnm estimate using gwa is [0.6600230500097711, 0.16728663936008945, 0.02512248420514078] , with standard errors [0.14248815 0.02447161 0.00488879] computed in  37.52255415916443 seconds.
    dnm estimate using gwasa is [0.6604981297820195, 0.16924607541398484, 0.02492054535741541] , with standard errors [0.14244908 0.02485465 0.00488222] computed in  0.8699958324432373 seconds.
    dnm estimate using ilt is [0.7466254648849691, 0.21415145383850764, 0.022794996238547492] , with standard errors [0.10187377 0.03137803        nan] computed in  1185.0924031734467 seconds.
    dnm estimate using oua is [0.5000083585920406, 0.5, 0.05] , with standard errors [       nan        nan 0.01961143] computed in  3.466001272201538 seconds.
    dnm estimate using uniform is [0.7477293759434092, 0.215047068344254, 0.022745437226772615] , with standard errors [0.16900378 0.03443071 0.00433504] computed in  3.275972366333008 seconds.

    How well methods perform varies from case to case. In this instance most methods perform well,
    while some throw errors but return useful output regardless, and some fail altogether.

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """

    pre_ll_fun = bld_ll_fun(data, likelihood, model, z_trunc, options)

    def error_fun(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return -pre_ll_fun(param)

    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll_fun(param)

    opt = ut.minimize_(error_fun, p0, p_bounds, con, opt_method, options)

    if p0.size > 1:
        try:
            cov = np.linalg.inv(-ut.Hessian(ll, opt.x, p_bounds))
        except:
            cov = 'Covariance matrix could not be determined.'
    else:
        try:
            cov = -1/ut.Hessian(ll, opt.x, p_bounds)
        except:
            cov = 'Covariance matrix could not be determined.'

    return opt, cov


def continuous_est_dnm(t_data, p_data, p0, b_rate, d_rate, p_bounds, con,
                       known_p, idx_known_p, opt_method, options):
    """Parameter estimation for continuously  observed continuous-time
    birth-and-death processes using direct numerical maximization of
    the likelihood.
    See :ref:`here <Continuously Observed Data>` for more information.

    To use this function call :func:`birdepy.estimate` with `scheme` set to
    'continuous'::

        birdepy.estimate(t_data, p_data, p0, p_bounds, scheme='continuous')


    Examples
    --------
    Simulate a continuous sample path and estimate the parameters.

    Import BirDePy:

    >>> import birdepy as bd

    Simulate some synthetic data:

    >>> t_data, p_data = bd.simulate.continuous([0.75, 0.25, 0.02, 1], 'Ricker', 10,
    ...                                         100, survival=True, seed=2021)

    Estimate:

    >>> est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
    ...                   model='Ricker', idx_known_p=[3], known_p=[1], scheme='continuous')
    >>> print(est.p)
    [0.7603171062895576, 0.2514810854871476, 0.020294342655751033]

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    --------


    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """

    if type(t_data[0]) != list:
        t_data = [t_data]
    if type(p_data[0]) != list:
        p_data = [p_data]

    holding_times = []
    for idx in range(len(t_data)):
        holding_times.append(np.array(t_data[idx][1:]) -
                             np.array(t_data[idx][0:-1]))

    flattened_p_data = [item for sublist in p_data for item in sublist]

    states_visited = np.arange(np.amin(flattened_p_data),
                               np.amax(flattened_p_data) + 1, 1)
    holding_t_data = np.zeros(len(states_visited))
    downward_jump_counts = np.zeros(len(states_visited))
    upward_jump_counts = np.zeros(len(states_visited))

    for idx1 in range(len(t_data)):
        state_record_temp = np.array(p_data[idx1])
        holding_times_temp = np.array(holding_times[idx1])
        for idx2, state in enumerate(states_visited):
            holding_t_data[idx2] += \
                np.sum(holding_times_temp[state_record_temp[:-1] == state])
            downward_jump_counts[idx2] += \
                np.sum(np.less(state_record_temp[1:], state_record_temp[0:-1])
                       & (state_record_temp[0:-1] == state))
            upward_jump_counts[idx2] += \
                np.sum(np.greater(state_record_temp[1:], state_record_temp[0:-1])
                       & (state_record_temp[0:-1] == state))


    def pre_ll(param):
        ll_ = 0
        for idx_, state_ in enumerate(states_visited):
            if b_rate(state_, param) > 0:
                ll_ += np.log(b_rate(state_, param)) * upward_jump_counts[idx_]
            if d_rate(state_, param) > 0:
                ll_ += np.log(d_rate(state_, param)) * downward_jump_counts[idx_]
            ll_ -= (b_rate(state_, param) + d_rate(state_, param)) * holding_t_data[idx_]
        return ll_

    def error_fun(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return -pre_ll(param)

    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    opt = ut.minimize_(error_fun, p0, p_bounds, con, opt_method, options)

    if p0.size > 1:
        try:
            cov = np.linalg.inv(-ut.Hessian(ll, opt.x, p_bounds))
        except:
            cov = 'Covariance matrix could not be determined.'
    else:
        try:
            cov = -1/ut.Hessian(ll, opt.x, p_bounds)
        except:
            cov = 'Covariance matrix could not be determined.'

    return opt, cov



from __future__ import print_function, absolute_import
import numpy as np
import birdepy.utility as ut
from numba import cuda
import math
from numba.cuda.random import create_xoroshiro128p_states
from numba.cuda.random import xoroshiro128p_uniform_float64 as rand
from collections import Counter


def discrete(param, model, z0, t, k=1, survival=False, seed=1):
    """Simulation of continuous-time birth-and-death processes at time 't'
    using CUDA.

    Parameters
    ----------
    param : array_like
        The parameters governing the evolution of the birth-and-death
        process to be simulated.
        Array of real elements of size (m,), where ‘m’ is the number of
        parameters.

    model : string
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
         Custom models are not available for this function. See :func:`birdepy.simulate.discrete()`
         for custom models.

    z0: int or callable
        The initial population size for each sample path.

    t : scalar
        The time at which the simulated birth-and-death is observed.

    k : int, optional
        The number of sample paths to be simulated.

    survival : bool, optional
        If set to True, then the simulated sample paths are conditioned to
        have a positive population size at the final observation time.  Since This
        can greatly increase computation time.

    seed : int, optional
        Seed for simulation.

    Return
    -------
    out : array_like
        A list containing sampled population size observations at
        time `t`, generated according to `model`.

    Examples
    --------
    Simulating 10 ** 8 sample paths of an M/M/inf queue with
    service rates 0.4 and arrival rate 0.2, with 10 items initially in the queue,
    observed at time 1.0:

    >>> from birdepy import gpu_functions as bdg
    >>> bdg.discrete([0.2, 0.4], 'M/M/inf', 10, 1.0, k=10**8)
    array([8, 6, 3, ..., 8, 7, 9], dtype=int64)

    Notes
    -----
    This function requires a compatible Nvidia graphics processing unit and
    drivers to be installed.

    The packages `Numba` and `cudatoolkit` also need to be installed.

    If you use this function for published work, then please cite [1].

    Sample paths are generated using a discrete-event simulation algorithm.
    See, for example, Algorithm 5.8 in [2].

    For a text book treatment on the theory of birth-and-death processes
    see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Kroese, D.P., Taimre, T. and Botev, Z.I. (2013) Handbook of Monte
     Carlo methods. John Wiley & Sons.

    .. [3] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """
    param = tuple(param)
    threads_per_block = 1024
    blocks = int(1 + k / threads_per_block)
    rng_states = create_xoroshiro128p_states(threads_per_block * blocks, seed=seed)
    out = np.zeros(threads_per_block * blocks, dtype=np.int64)
    if model == 'Verhulst':
        discrete_verhulst[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Ricker':
        discrete_ricker[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Hassell':
        discrete_hassell[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'MS-S':
        discrete_mss[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Moran':
        discrete_moran[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'pure-birth':
        discrete_pb[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'pure-death':
        discrete_pd[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Poisson':
        discrete_poisson[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'linear':
        discrete_linear[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'linear-migration':
        discrete_lm[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'M/M/1':
        discrete_mm1[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'M/M/inf':
        discrete_mminf[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'loss-system':
        discrete_loss[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    else:
        raise TypeError("Argument 'model' has an unknown value.")


def probability(z0, zt, t, param, model, k=10**6, seed=1):
    """Transition probabilities for continuous-time birth-and-death processes
    generated using Monte Carlo on a GPU.

    Parameters
    ----------
    z0 : array_like
        States of birth-and-death process at time 0.

    zt : array_like
        States of birth-and-death process at time(s) `t`.

    t : array_like
        Elapsed time(s) (if has size greater than 1, then must be increasing).

    param : array_like
        The parameters governing the evolution of the birth-and-death
        process.
        Array of real elements of size (n,), where ‘n’ is the number of
        parameters.

    model : string, optional
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'MS-S'
            - 'M/M/inf'
            - 'loss-system'
         Custom models are not available for this function. See :func:`birdepy.probability()`
         for custom models.

    k: int, optional
        Minimum number of samples used to generate each probability estimate.
        (Actual number of samples will usually be higher due to the way memory
        is allocated on GPU.)
        The total number of samples used will be at least z0.size * k.

    seed : int, optional
        Seed for simulations.

    Return
    -------
    transition_probability : ndarray
        An array of transition probabilities. If t has size bigger than 1,
        then the first coordinate corresponds to `t`, the second coordinate
        corresponds to `z0` and the third coordinate corresponds to `zt`;
        for example if `z0=[1,3,5,10]`, `zt=[5,8]` and `t=[1,2,3]`, then
        `transition_probability[2,0,1]` corresponds to
        P(Z(3)=8 | Z(0)=1).
        If `t` has size 1 the first coordinate corresponds to `z0` and the second
        coordinate corresponds to `zt`.

    Examples
    --------
    >>> from birdepy import gpu_functions as bdg
    ... param = (210, 20, 0.002, 0, 100)
    ... t = 0.2
    ... z0 = [50, 60]
    ... zt = [55, 56, 57, 58, 59,60]
    ... bdg.probability(z0, zt, t, param, 'Moran', 10**6)
    array([[3.09160e-02, 5.43120e-02, 8.09760e-02, 1.05968e-01, 1.23203e-01,1.27453e-01],
       [0.00000e+00, 0.00000e+00, 0.00000e+00, 0.00000e+00, 0.00000e+00, 6.30000e-05]])

    Notes
    -----
    This function requires a compatible Nvidia graphics processing unit and
    drivers to be installed.

    The packages `Numba` and `cudatoolkit` also need to be installed.

    If you use this function for published work, then please cite [1].

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`


    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    param = tuple(param)
    if np.isscalar(z0):
        z0 = np.array([z0])
    else:
        z0 = np.array(z0)
    if np.isscalar(zt):
        zt = np.array([zt])
    else:
        zt = np.array(zt)
    if np.isscalar(t):
        t = np.array([t])
    else:
        t = np.array(t)

    if t.size == 1:
        output = np.zeros((z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            sim = discrete(param, model, _z0, t[0], k, False, seed)
            counts = Counter(sim)
            for idx2, _zt in enumerate(zt):
                output[idx1, idx2] = counts[_zt]/k
    else:
        output = np.zeros((t.size, z0.size, zt.size))
        for idx3, _t in enumerate(t):
            for idx1, _z0 in enumerate(z0):
                sim = discrete(param, model, _z0, _t, k, False, seed)
                counts = Counter(sim)
                for idx2, _zt in enumerate(zt):
                    output[idx3, idx1, idx2] = counts[_zt] / k
    return output


@cuda.jit
def discrete_verhulst(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * (1 - p[2]*z) * z

    def d_rate(z):
        return p[1] * (1 + p[3]*z) * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_ricker(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * math.exp(-(p[2] * z) ** p[3]) * z

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_bh(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * p[2] / (z + p[2])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_hassell(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z / (1 + z * p[2]) ** p[3]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mss(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z / (1 + (z * p[2]) ** p[3])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_moran(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return (p[4] - z) * (p[0] * z * (1 - p[2]) / p[4] +
                                     p[1] * (p[4] - z) * p[3] / p[4]) \
                       / p[4]

    def d_rate(z):
        return z * (p[2] * (p[4] - z) * (1 - p[3]) / p[4] +
                            p[0] * z * p[2] / p[4]) \
                       / p[4]

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_pb(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z

    def d_rate(z):
        return 0

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_pd(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return 0

    def d_rate(z):
        return p[0] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_poisson(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return 0

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop



@cuda.jit
def discrete_linear(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_lm(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z + p[2]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mm1(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return p[1] * (z > 0)

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mminf(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_loss(out, rng_states, p, z0, time, survival):
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * (z < p[2])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld
from birdepy.interface_dnm import bld_ll_fun
import birdepy.utility_em as utem


def discrete_est_em_cg(data, sorted_data, p0, likelihood, technique, known_p,
                       idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
                       con, max_it, i_tol, j_tol, h_tol, display, opt_method,
                       options):

    iterations = [p0]

    pre_ll = bld_ll_fun(data, likelihood, model, z_trunc, options)

    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    it = 0
    p_est_it_m1 = p0
    difference = i_tol + 1
    while difference > i_tol and it < max_it:

        f_fun = f_fun_bld(sorted_data, p_est_it_m1, b_rate, d_rate, likelihood,
                          technique, idx_known_p, known_p, model, z_trunc,
                          j_tol, h_tol, options)[0]

        def error_fun(p_prop):
            return -f_fun(p_prop)

        g_tilde = ut.minimize_(error_fun, p_est_it_m1, p_bounds, con,
                               opt_method, options).x - p_est_it_m1

        if (it % p0.size) == 0:
            direction = g_tilde
        else:
            J_ell_it_m1 = ut.Jacobian(ll, p_est_it_m1, p_bounds)
            J_ell_it_m2 = ut.Jacobian(ll, p_est_it_m2, p_bounds)

            previous_direction = direction

            beta = np.divide(np.inner(g_tilde,
                                      np.subtract(J_ell_it_m1, J_ell_it_m2)),
                             np.inner(previous_direction,
                                      np.subtract(J_ell_it_m1, J_ell_it_m2)))
            direction = g_tilde - np.multiply(beta, previous_direction)

        p_est_it_m2 = p_est_it_m1

        a_opt = utem.line_search_ll(direction, data, p_est_it_m1, likelihood,
                                    known_p, idx_known_p, model, z_trunc,
                                    p_bounds, con, opt_method, options)[0]

        p_est_it_m1 = p_est_it_m1 + np.multiply(a_opt, direction)

        difference = np.sum(np.abs(p_est_it_m1 - p_est_it_m2))
        iterations.append(p_est_it_m1)

        it += 1
        if display:
            print('Iteration ', it, ' estimate is: ', p_est_it_m1)


    return list(p_est_it_m1), np.array(iterations)
